//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                         examples/concept-based_1.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows a C++03 compliant example of static specialization with the C4TS++ library (*).
//
// (*) Bruno Bachelet, Antoine Mahul and Loic Yon.
//     Template Metaprogramming Techniques for Concept-Based Specialization.
//     Research report LIMOS/RR10-18, LIMOS, UMR 6158-CNRS,
//     Universite Blaise Pascal, BP 10125, 63173 Aubiere, France.

// Headers //---------------------------------------------------------------------------------------
#include <list>
#include <map>
#include <set>
#include <vector>
#include <gnx_concept.hpp>

// Serialization Code //----------------------------------------------------------------------------
#define C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
#include <serialization.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------

//------------------------------------------------------------------------------Concepts Declaration
gnx_declare_concept(SingleObject);
gnx_declare_concept(ComplexObject);
gnx_declare_concept(STLPair);
gnx_declare_concept(STLContainer);
gnx_declare_concept(STLSequence);
gnx_declare_concept(STLAssociative);

//---------------------------------------------------------------Modeling & Refinement Relationships
template <> struct gnx_models_concept<char,SingleObject> : gnx_true {};
template <> struct gnx_models_concept<int,SingleObject> : gnx_true {};
template <> struct gnx_models_concept<float,SingleObject> : gnx_true {};

template <class T1,class T2>
struct gnx_models_concept<std::pair<T1,T2>,STLPair> : gnx_true {};

template <class T>
struct gnx_models_concept<std::vector<T>,STLSequence> : gnx_true {};

template <class T>
struct gnx_models_concept<std::list<T>,STLSequence> : gnx_true {};

template <class T>
struct gnx_models_concept<std::set<T>,STLAssociative> : gnx_true {};

template <class K,class T>
struct gnx_models_concept<std::map<K,T>,STLAssociative> : gnx_true {};

template <>
struct gnx_models_concept<STLPair,ComplexObject> : gnx_true {};

template <>
struct gnx_models_concept<STLContainer,ComplexObject> : gnx_true {};

template <>
struct gnx_models_concept<STLSequence,STLContainer> : gnx_true {};

template <>
struct gnx_models_concept<STLAssociative,STLContainer> : gnx_true {};

// Serializer Class //------------------------------------------------------------------------------
template <class T> struct ErrorSpecializationNotFound;

struct SerializerContext;

template <class T,class = gnx_best_concept(SerializerContext,T)>
class Serializer : ErrorSpecializationNotFound<T> {};

//-----------------------------------------------------------------------------Single Object Version
template <> struct gnx_uses_concept<SerializerContext,SingleObject> : gnx_true {};

template <class T> class Serializer<T,SingleObject> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_single_object(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_single_object(object,copy,display); }
};

//----------------------------------------------------------------------------------STL Pair Version
template <> struct gnx_uses_concept<SerializerContext,STLPair> : gnx_true {};

template <class T> class Serializer<T,STLPair> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_pair(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_pair(object,copy,display); }
};

//-----------------------------------------------------------------------------STL Container Version
template <> struct gnx_uses_concept<SerializerContext,STLContainer> : gnx_true {};

template <class T> class Serializer<T,STLContainer> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_container(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_container(object,copy,display); }
};

//------------------------------------------------------------------------------STL Sequence Version
template <> struct gnx_uses_concept<SerializerContext,STLSequence> : gnx_true {};

template <class T> class Serializer<T,STLSequence> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_sequence(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_sequence(object,copy,display); }
};

//---------------------------------------------------------------------------STL Associative Version
template <> struct gnx_uses_concept<SerializerContext,STLAssociative> : gnx_true {};

template <class T> class Serializer<T,STLAssociative> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

// Serialization Code //----------------------------------------------------------------------------
#include <serialization.hpp>

// Testing Code //----------------------------------------------------------------------------------
#include <testing.hpp>

// End //-------------------------------------------------------------------------------------------
