//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                                examples/dispatch.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows an example of static specialization with concept-based dispatch (*).
//
// (*) Brian McNamara and Yannis Smaragdakis.
//     Static Interfaces in C++.
//     In First Workshop on C++ Template Programming, 2000.

// Headers //---------------------------------------------------------------------------------------
#include <list>
#include <map>
#include <set>
#include <vector>

// Serialization Code //----------------------------------------------------------------------------
#define C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
#include <serialization.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------
struct SingleObject;
struct STLPair;
struct STLContainer;
struct STLSequence;
struct STLAssociative;

template <class T,class C> struct StaticIsA { static const bool valid = false; };

template <> struct StaticIsA<char,SingleObject> { static const bool valid = true; };
template <> struct StaticIsA<int,SingleObject> { static const bool valid = true; };
template <> struct StaticIsA<float,SingleObject> { static const bool valid = true; };

template <class T1,class T2>
struct StaticIsA< std::pair<T1,T2>,STLPair > { static const bool valid = true; };

template <class T,class ALLOC>
struct StaticIsA< std::vector<T,ALLOC>,STLContainer > { static const bool valid = true; };

template <class T,class ALLOC>
struct StaticIsA< std::list<T,ALLOC>,STLContainer > { static const bool valid = true; };

template <class T,class ALLOC,class COMP>
struct StaticIsA< std::set<T,COMP,ALLOC>,STLContainer > { static const bool valid = true; };

template <class T,class K,class ALLOC,class COMP>
struct StaticIsA< std::map<T,K,COMP,ALLOC>,STLContainer > { static const bool valid = true; };

template <class T,class ALLOC>
struct StaticIsA< std::vector<T,ALLOC>,STLSequence > { static const bool valid = true; };

template <class T,class ALLOC>
struct StaticIsA< std::list<T,ALLOC>,STLSequence > { static const bool valid = true; };

template <class T,class ALLOC,class COMP>
struct StaticIsA< std::set<T,COMP,ALLOC>,STLAssociative > { static const bool valid = true; };

template <class T,class K,class ALLOC,class COMP>
struct StaticIsA< std::map<T,K,COMP,ALLOC>,STLAssociative > { static const bool valid = true; };

// No difference between non-constant and constant types. //
template <class TYPE,class CONCEPT>
struct StaticIsA<const TYPE,CONCEPT> : StaticIsA<TYPE,CONCEPT> {};

// Dispatcher //------------------------------------------------------------------------------------
enum { IS_SINGLE_OBJECT,IS_STL_PAIR,IS_STL_CONTAINER,
       IS_STL_SEQUENCE,IS_STL_ASSOCIATIVE,UNSPECIFIED };

template <class T> struct Dispatcher {
 static const int which = StaticIsA<T,STLAssociative>::valid ? IS_STL_ASSOCIATIVE
                          : StaticIsA<T,STLSequence>::valid ? IS_STL_SEQUENCE
                          : StaticIsA<T,STLContainer>::valid ? IS_STL_CONTAINER
                          : StaticIsA<T,STLPair>::valid ? IS_STL_PAIR
                          : StaticIsA<T,SingleObject>::valid ? IS_SINGLE_OBJECT
                          : UNSPECIFIED;
};

// Serializer Class //------------------------------------------------------------------------------
template <class T> struct ErrorSpecializationNotFound;

template <class T,int = Dispatcher<T>::which>
class Serializer : ErrorSpecializationNotFound<T> {};

//-----------------------------------------------------------------------------Single Object Version
template <class T> class Serializer<T,IS_SINGLE_OBJECT> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_single_object(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_single_object(object,copy,display); }
};

//----------------------------------------------------------------------------------STL Pair Version
template <class T> class Serializer<T,IS_STL_PAIR> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_pair(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_pair(object,copy,display); }
};

//-----------------------------------------------------------------------------STL Container Version
template <class T> class Serializer<T,IS_STL_CONTAINER> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_container(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_container(object,copy,display); }
};

//------------------------------------------------------------------------------STL Sequence Version
template <class T> class Serializer<T,IS_STL_SEQUENCE> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_sequence(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_sequence(object,copy,display); }
};

//---------------------------------------------------------------------------STL Associative Version
template <class T> class Serializer<T,IS_STL_ASSOCIATIVE> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

// Serialization Code //----------------------------------------------------------------------------
#include <serialization.hpp>

// Testing Code //----------------------------------------------------------------------------------
#include <testing.hpp>

// End //-------------------------------------------------------------------------------------------
