//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                               examples/enable_if.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows an example of static specialization with concept-based overloading (*).
//
// (*) Jaakko Jarvi, Jeremiah Willcock, and Andrew Lumsdaine.
//     Concept-Controlled Polymorphism.
//     In Lecture Notes in Computer Science, volume 2830, pages 228--244. Springer-Verlag, 2003.

// Headers //---------------------------------------------------------------------------------------
#include <list>
#include <map>
#include <set>
#include <vector>

// Serialization Code //----------------------------------------------------------------------------
#define C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
#include <serialization.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------
template <class T> struct is_SingleObject { static const bool value = false; };
template <class T> struct is_STLPair { static const bool value = false; };
template <class T> struct is_STLContainer { static const bool value = false; };
template <class T> struct is_STLSequence { static const bool value = false; };
template <class T> struct is_STLAssociative { static const bool value = false; };

template <> struct is_SingleObject<char> { static const bool value = true; };
template <> struct is_SingleObject<int> { static const bool value = true; };
template <> struct is_SingleObject<float> { static const bool value = true; };

template <class T1,class T2>
struct is_STLPair< std::pair<T1,T2> > { static const bool value = true; };

/* Try to uncomment the following code.
   It causes ambiguities in the static specialization of "Serializer". */

/*
template <class T,class ALLOC>
struct is_STLContainer< std::vector<T,ALLOC> > { static const bool value = true; };

template <class T,class ALLOC>
struct is_STLContainer< std::list<T,ALLOC> > { static const bool value = true; };

template <class T,class ALLOC,class COMP>
struct is_STLContainer< std::set<T,COMP,ALLOC> > { static const bool value = true; };

template <class T,class K,class ALLOC,class COMP>
struct is_STLContainer< std::map<T,K,COMP,ALLOC> > { static const bool value = true; };
*/

template <class T,class ALLOC>
struct is_STLSequence< std::vector<T,ALLOC> > { static const bool value = true; };

template <class T,class ALLOC>
struct is_STLSequence< std::list<T,ALLOC> > { static const bool value = true; };

template <class T,class ALLOC,class COMP>
struct is_STLAssociative< std::set<T,COMP,ALLOC> > { static const bool value = true; };

template <class T,class K,class ALLOC,class COMP>
struct is_STLAssociative< std::map<T,K,COMP,ALLOC> > { static const bool value = true; };

// No difference between non-constant and constant types. //
template <class T> struct is_SingleObject<const T> : is_SingleObject<T> {};
template <class T> struct is_STLPair<const T> : is_STLPair<T> {};
template <class T> struct is_STLContainer<const T> : is_STLContainer<T> {};
template <class T> struct is_STLSequence<const T> : is_STLSequence<T> {};
template <class T> struct is_STLAssociative<const T> : is_STLAssociative<T> {};

// "enable_if" Metafunction //----------------------------------------------------------------------
template <bool B,class T = void>
struct enable_if_c { typedef T type; };

template <class T>
struct enable_if_c<false,T> {};

template <class COND,class T = void>
struct enable_if : enable_if_c<COND::value,T> {};

// Serializer Class //------------------------------------------------------------------------------
template <class T> struct ErrorSpecializationNotFound;

template <class T,class = void>
class Serializer : ErrorSpecializationNotFound<T> {};

//-----------------------------------------------------------------------------Single Object Version
template <class T>
class Serializer<T,typename enable_if< is_SingleObject<T> >::type> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_single_object(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_single_object(object,copy,display); }
};

//----------------------------------------------------------------------------------STL Pair Version
template <class T>
class Serializer<T,typename enable_if< is_STLPair<T> >::type> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_pair(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_pair(object,copy,display); }
};

//-----------------------------------------------------------------------------STL Container Version
template <class T>
class Serializer<T,typename enable_if< is_STLContainer<T> >::type> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_container(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_container(object,copy,display); }
};

//------------------------------------------------------------------------------STL Sequence Version
template <class T>
class Serializer<T,typename enable_if< is_STLSequence<T> >::type> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_sequence(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_sequence(object,copy,display); }
};

//---------------------------------------------------------------------------STL Associative Version
template <class T>
class Serializer<T,typename enable_if< is_STLAssociative<T> >::type> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

// Serialization Code //----------------------------------------------------------------------------
#include <serialization.hpp>

// Testing Code //----------------------------------------------------------------------------------
#include <testing.hpp>

// End //-------------------------------------------------------------------------------------------
