//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                           examples/serialization.hpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file contains the serialization code used in the different examples.

// Guardian //--------------------------------------------------------------------------------------
#ifndef _C4TS_EXAMPLES_SERIALIZATION_HPP_
#define _C4TS_EXAMPLES_SERIALIZATION_HPP_

// Headers //---------------------------------------------------------------------------------------
#include <iostream>

// Functions Prototypes //--------------------------------------------------------------------------
#ifdef C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
 template <class T> int deflate_single_object(char *,const T &,bool = false);
 template <class T> int inflate_single_object(T &,const char *,bool = false);

 template <class T> int deflate_stl_pair(char *,const T &,bool = false);
 template <class T> int inflate_stl_pair(T &,const char *,bool = false);

 template <class T> int deflate_stl_container(char *,const T &,bool = false);
 template <class T> int inflate_stl_container(T &,const char *,bool = false);

 template <class T> int deflate_stl_sequence(char *,const T &,bool = false);
 template <class T> int inflate_stl_sequence(T &,const char *,bool = false);

 template <class T> int deflate_stl_associative(char *,const T &,bool = false);
 template <class T> int inflate_stl_associative(T &,const char *,bool = false);

 #undef C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
 #undef _C4TS_EXAMPLES_SERIALIZATION_HPP_
#else

// Single Object Serialization //-------------------------------------------------------------------

//-------------------------------------------------------------------------------------------Deflate
template <class T>
int deflate_single_object(char * copy,const T & object,bool display) {
 static const int n = sizeof(T);

 if (display) std::cout << "    Deflating single object..." << std::endl;

 const char * mem = (const char *)(&object);

 for (int i=0; i<n; ++i) copy[i]=mem[i];
 return n;
}
//-------------------------------------------------------------------------------------------Inflate
template <class T>
int inflate_single_object(T & object,const char * copy,bool display) {
 static const int n = sizeof(T);

 if (display) std::cout << "    Inflating single object..." << std::endl;

 char * mem = (char *)(&object);

 for (int i=0; i<n; ++i) mem[i]=copy[i];
 return n;
}

// STL Pair Serialization //------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------Deflate
template <class T>
int deflate_stl_pair(char * copy,const T & pair,bool display) {
 if (display) std::cout << "    Deflating STL pair..." << std::endl;

 char * current = copy;

 current+=Serializer<typename T::first_type>::deflate(current,pair.first);
 current+=Serializer<typename T::second_type>::deflate(current,pair.second);
 return current-copy;
}

//-------------------------------------------------------------------------------------------Inflate
template <class T>
int inflate_stl_pair(T & pair,const char * copy,bool display) {
 if (display) std::cout << "    Inflating STL pair..." << std::endl;

 const char * current = copy;

 current+=Serializer<typename T::first_type>::inflate(pair.first,current);
 current+=Serializer<typename T::second_type>::inflate(pair.second,current);
 return current-copy;
}

// STL Container Serialization //-------------------------------------------------------------------

//-------------------------------------------------------------------------------------------Deflate
template <class T>
int deflate_stl_container(char * copy,const T & container,bool display) {
 if (display) std::cout << "    Deflating STL container..." << std::endl;

 char * current = copy;

 current+=Serializer<int>::deflate(current,int(container.size()));

 for (typename T::const_iterator it = container.begin(); it!=container.end(); ++it)
  current+=Serializer<typename T::value_type>::deflate(current,*it);

 return current-copy;
}

//-------------------------------------------------------------------------------------------Inflate
template <class T>
int inflate_stl_container(T & container,const char * copy,bool display) {
 if (display) std::cout << "    Inflating STL container..." << std::endl;

 const char * current = copy;
 int size;

 container.clear();
 current+=Serializer<int>::inflate(size,copy);

 for (int i = 0; i<size; ++i) {
  typename T::value_type object;

  current+=Serializer<typename T::value_type>::inflate(object,current);
  container.insert(container.end(),object);
 }

 return current-copy;
}

// STL Sequence Serialization //--------------------------------------------------------------------

//-------------------------------------------------------------------------------------------Deflate
template <class T>
int deflate_stl_sequence(char * copy,const T & container,bool display) {
 if (display) std::cout << "    Deflating STL sequence..." << std::endl;

 return deflate_stl_container(copy,container);
}

//-------------------------------------------------------------------------------------------Inflate
template <class T>
int inflate_stl_sequence(T & container,const char * copy,bool display) {
 if (display) std::cout << "    Inflating STL sequence..." << std::endl;

 const char * current = copy;
 int size;

 container.clear();
 current+=Serializer<int>::inflate(size,copy);

 for (int i = 0; i<size; ++i) {
  typename T::value_type object;

  current+=Serializer<typename T::value_type>::inflate(object,current);
  container.push_back(object);
 }

 return current-copy;
}

// STL Associative Serialization //-----------------------------------------------------------------

//-------------------------------------------------------------------------------------------Deflate
template <class T>
int deflate_stl_associative(char * copy,const T & container,bool display) {
 if (display) std::cout << "    Deflating STL associative..." << std::endl;

 return deflate_stl_container(copy,container);
}

//-------------------------------------------------------------------------------------------Inflate
template <class T>
int inflate_stl_associative(T & container,const char * copy,bool display) {
 if (display) std::cout << "    Inflating STL associative..." << std::endl;

 const char * current = copy;
 int size;

 container.clear();
 current+=Serializer<int>::inflate(size,copy);

 for (int i = 0; i<size; ++i) {
  typename T::value_type object;

  current+=Serializer<typename T::value_type>::inflate(object,current);
  container.insert(object);
 }

 return current-copy;
}

// End //-------------------------------------------------------------------------------------------
#endif
#endif
