//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                         examples/concept-based_1.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows a C++11 compliant example of static specialization with the C4TS++ library (*).
//
// (*) Bruno Bachelet, Antoine Mahul and Loic Yon.
//     Template Metaprogramming Techniques for Concept-Based Specialization.
//     Research report LIMOS/RR10-18, LIMOS, UMR 6158-CNRS,
//     Universite Blaise Pascal, BP 10125, 63173 Aubiere, France.

// Headers //---------------------------------------------------------------------------------------
#include <list>
#include <map>
#include <set>
#include <vector>
#include <gnx_concept.hpp>

// Serialization Code //----------------------------------------------------------------------------
#define C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
#include <serialization.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------

//------------------------------------------------------------------------------Concepts Declaration
gnx_declare_concept(SingleObject);
gnx_declare_concept(ComplexObject);
gnx_declare_concept(STLPair);
gnx_declare_concept(STLContainer);
gnx_declare_concept(STLSequence);
gnx_declare_concept(STLAssociative);

//---------------------------------------------------------------Modeling & Refinement Relationships
template <> gnx_add_models(char,SingleObject);
template <> gnx_add_models(int,SingleObject);
template <> gnx_add_models(float,SingleObject);

template <class T1,class T2> gnx_add_models(std::pair<T1,T2>,STLPair);
template <class T>           gnx_add_models(std::vector<T>,STLSequence);
template <class T>           gnx_add_models(std::list<T>,STLSequence);
template <class T>           gnx_add_models(std::set<T>,STLAssociative);
template <class K,class T>   gnx_add_models(std::map<K,T>,STLAssociative);

template <> gnx_add_models(STLPair,ComplexObject);
template <> gnx_add_models(STLContainer,ComplexObject);
template <> gnx_add_models(STLSequence,STLContainer);
template <> gnx_add_models(STLAssociative,STLContainer);

// Serializer Class //------------------------------------------------------------------------------
template <class T> struct ErrorSpecializationNotFound;

struct SerializerContext;

template <class T,
          class = gnx_best_concept_t<SerializerContext,T>
         >
class Serializer : ErrorSpecializationNotFound<T> {};

//-----------------------------------------------------------------------------Single Object Version
template <> gnx_add_uses(SerializerContext,SingleObject);

template <class T> class Serializer<T,SingleObject> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_single_object(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_single_object(object,copy,display); }
};

//----------------------------------------------------------------------------------STL Pair Version
template <> gnx_add_uses(SerializerContext,STLPair);

template <class T> class Serializer<T,STLPair> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_pair(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_pair(object,copy,display); }
};

//-----------------------------------------------------------------------------STL Container Version
template <> gnx_add_uses(SerializerContext,STLContainer);

template <class T> class Serializer<T,STLContainer> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_container(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_container(object,copy,display); }
};

//------------------------------------------------------------------------------STL Sequence Version
template <> gnx_add_uses(SerializerContext,STLSequence);

template <class T> class Serializer<T,STLSequence> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_sequence(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_sequence(object,copy,display); }
};

//---------------------------------------------------------------------------STL Associative Version
template <> gnx_add_uses(SerializerContext,STLAssociative);

template <class T> class Serializer<T,STLAssociative> {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

// Serialization Code //----------------------------------------------------------------------------
#include <serialization.hpp>

// Testing Code //----------------------------------------------------------------------------------
#include <testing.hpp>

// End //-------------------------------------------------------------------------------------------
