//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                         examples/concept-based_3.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows a C++11 compliant example of static specialization with the C4TS++ library (*).
//
// (*) Bruno Bachelet, Antoine Mahul and Loic Yon.
//     Template Metaprogramming Techniques for Concept-Based Specialization.
//     Research report LIMOS/RR10-18, LIMOS, UMR 6158-CNRS,
//     Universite Blaise Pascal, BP 10125, 63173 Aubiere, France.

// Headers //---------------------------------------------------------------------------------------
#include <iostream>
#include <string>
#include <typeinfo>
#include <gnx_concept.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------

//------------------------------------------------------------------------------Concepts Declaration
gnx_declare_concept(Concept1);
gnx_declare_concept(Concept2);
gnx_declare_concept(Concept3);
gnx_declare_concept(Concept4);
gnx_declare_concept(Concept5);
gnx_index_concept(gnx_cor<Concept1,Concept2>);
gnx_index_concept(gnx_cand<Concept1,Concept2>);
gnx_index_concept(gnx_cnot<Concept3>);

//-------------------------------------------------------------------------------Classes Declaration
class ClassA {};
class ClassB {};
class ClassC {};
class ClassD {};
class ClassE {};

//---------------------------------------------------------------Modeling & Refinement Relationships
template <> gnx_add_models(ClassA,Concept1);
template <> gnx_add_models(ClassB,Concept5);
template <> gnx_add_models(ClassC,Concept1);
template <> gnx_add_models(ClassC,Concept5);
template <> gnx_add_models(ClassD,Concept3);
template <> gnx_add_models(ClassE,Concept4);

template <> gnx_add_models(Concept5,Concept2);

// Trait Class //-----------------------------------------------------------------------------------
struct TraitContext;

template <class TYPE,
          class = gnx_best_concept_t<TraitContext,TYPE>
         >
struct TraitClass {
 static std::string value(void) { return "unknown"; }
};

//----------------------------------------------------------------------------------Concept1 Version
template <> gnx_add_uses(TraitContext,Concept1);

template <class TYPE> struct TraitClass<TYPE,Concept1> {
 static std::string value(void) { return "Concept1"; }
};

//----------------------------------------------------------------------------------Concept3 Version
template <> gnx_add_uses(TraitContext,Concept3);

template <class TYPE> struct TraitClass<TYPE,Concept3> {
 static std::string value(void) { return "Concept3"; }
};

//----------------------------------------------------------------------Concept1 && Concept2 Version
template <> gnx_add_uses(TraitContext,gnx_cand<Concept1,Concept2>);

template <class TYPE> struct TraitClass< TYPE,gnx_cand<Concept1,Concept2> > {
 static std::string value(void) { return "Concept1&&Concept2"; }
};

//----------------------------------------------------------------------Concept1 || Concept2 Version
template <> gnx_add_uses(TraitContext,gnx_cor<Concept1,Concept2>);

template <class TYPE> struct TraitClass< TYPE,gnx_cor<Concept1,Concept2> > {
 static std::string value(void) { return "Concept1||Concept2"; }
};

//---------------------------------------------------------------------------------!Concept3 Version
template <> gnx_add_uses(TraitContext,gnx_cnot<Concept3>);

template <class TYPE> struct TraitClass< TYPE,gnx_cnot<Concept3> > {
 static std::string value(void) { return "!Concept3"; }
};

// Test //------------------------------------------------------------------------------------------
int main(void) {
 std::cout << "A matches 1 ? " << gnx_matches_concept<ClassA,Concept1>::value << std::endl;
 std::cout << "A matches 2 ? " << gnx_matches_concept<ClassA,Concept2>::value << std::endl;
 std::cout << "A matches 3 ? " << gnx_matches_concept<ClassA,Concept3>::value << std::endl;
 std::cout << "A matches 4 ? " << gnx_matches_concept<ClassA,Concept4>::value << std::endl;
 std::cout << "A matches 5 ? " << gnx_matches_concept<ClassA,Concept5>::value << std::endl;
 std::cout << std::endl;

 std::cout << "B matches 1 ? " << gnx_matches_concept<ClassB,Concept1>::value << std::endl;
 std::cout << "B matches 2 ? " << gnx_matches_concept<ClassB,Concept2>::value << std::endl;
 std::cout << "B matches 3 ? " << gnx_matches_concept<ClassB,Concept3>::value << std::endl;
 std::cout << "B matches 4 ? " << gnx_matches_concept<ClassB,Concept4>::value << std::endl;
 std::cout << "B matches 5 ? " << gnx_matches_concept<ClassB,Concept5>::value << std::endl;
 std::cout << std::endl;

 std::cout << "C matches 1 ? " << gnx_matches_concept<ClassC,Concept1>::value << std::endl;
 std::cout << "C matches 2 ? " << gnx_matches_concept<ClassC,Concept2>::value << std::endl;
 std::cout << "C matches 3 ? " << gnx_matches_concept<ClassC,Concept3>::value << std::endl;
 std::cout << "C matches 4 ? " << gnx_matches_concept<ClassC,Concept4>::value << std::endl;
 std::cout << "C matches 5 ? " << gnx_matches_concept<ClassC,Concept5>::value << std::endl;
 std::cout << std::endl;

 std::cout << "D matches 1 ? " << gnx_matches_concept<ClassD,Concept1>::value << std::endl;
 std::cout << "D matches 2 ? " << gnx_matches_concept<ClassD,Concept2>::value << std::endl;
 std::cout << "D matches 3 ? " << gnx_matches_concept<ClassD,Concept3>::value << std::endl;
 std::cout << "D matches 4 ? " << gnx_matches_concept<ClassD,Concept4>::value << std::endl;
 std::cout << "D matches 5 ? " << gnx_matches_concept<ClassD,Concept5>::value << std::endl;
 std::cout << std::endl;

 std::cout << "E matches 1 ? " << gnx_matches_concept<ClassE,Concept1>::value << std::endl;
 std::cout << "E matches 2 ? " << gnx_matches_concept<ClassE,Concept2>::value << std::endl;
 std::cout << "E matches 3 ? " << gnx_matches_concept<ClassE,Concept3>::value << std::endl;
 std::cout << "E matches 4 ? " << gnx_matches_concept<ClassE,Concept4>::value << std::endl;
 std::cout << "E matches 5 ? " << gnx_matches_concept<ClassE,Concept5>::value << std::endl;
 std::cout << std::endl;

 typedef gnx_best_concept_t<TraitContext,ClassA> ConceptA;
 typedef gnx_best_concept_t<TraitContext,ClassB> ConceptB;
 typedef gnx_best_concept_t<TraitContext,ClassC> ConceptC;
 typedef gnx_best_concept_t<TraitContext,ClassD> ConceptD;
 typedef gnx_best_concept_t<TraitContext,ClassE> ConceptE;

 std::cout << "best_concept<A> = " << typeid(ConceptA).name() << std::endl;
 std::cout << "best_concept<B> = " << typeid(ConceptB).name() << std::endl;
 std::cout << "best_concept<C> = " << typeid(ConceptC).name() << std::endl;
 std::cout << "best_concept<D> = " << typeid(ConceptD).name() << std::endl;
 std::cout << "best_concept<E> = " << typeid(ConceptE).name() << std::endl;
 std::cout << std::endl;

 std::cout << "Trait<A> = " << TraitClass<ClassA>::value() << std::endl;
 std::cout << "Trait<B> = " << TraitClass<ClassB>::value() << std::endl;
 std::cout << "Trait<C> = " << TraitClass<ClassC>::value() << std::endl;
 std::cout << "Trait<D> = " << TraitClass<ClassD>::value() << std::endl;
 std::cout << "Trait<E> = " << TraitClass<ClassE>::value() << std::endl;
}

// End //-------------------------------------------------------------------------------------------
