//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                       c4ts++_optimized/gnx_basic.hpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This file is part of the C4TS++ library. This library is free software; you can redistribute it
// and/or modify it under the terms of the GNU Library General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
// the GNU Library General Public License for more details (http://www.gnu.org).
//
// This file provides basic metaprogramming elements.

// Guardian //--------------------------------------------------------------------------------------
#ifndef _C4TS_GNX_BASIC_
#define _C4TS_GNX_BASIC_

//-----------------------------------------------------------------------------------------gnx_cpp11
/* Macro that is defined only if the compilation performs according to the C++11 standard. */
#if __cplusplus >= 201103L
 #define gnx_cpp11
#else
 #undef gnx_cpp11
#endif

//------------------------------------------------------------------------------------------gnx_type
/* Makes a metadata from a type. */
template <class TYPE> struct gnx_type { typedef TYPE type; };

//-----------------------------------------------------------------------------------------gnx_value
/* Makes a metadata from a value. */
template <class TYPE,TYPE VALUE> struct gnx_value { static const TYPE value = VALUE; };

//------------------------------------------------------------------------------------------gnx_bool
/* Makes a metadata from a boolean value. */
template <bool VALUE> struct gnx_bool : gnx_value<bool,VALUE> {};

//-----------------------------------------------------------------------------------------gnx_false
/* Makes a metadata from the "false" value. */
typedef gnx_value<bool,false> gnx_false;

//------------------------------------------------------------------------------------------gnx_true
/* Makes a metadata from the "true" value. */
typedef gnx_value<bool,true> gnx_true;

//------------------------------------------------------------------------------------------gnx_same
/* Returns whether two types are identical. */
template <class TYPE1,class TYPE2> struct gnx_same : gnx_false {};

template <class TYPE> struct gnx_same<TYPE,TYPE> : gnx_true {};

//--------------------------------------------------------------------------------------------gnx_if
/* Chooses between two types based on a boolean metadata. */
template <class TEST,class IF,class ELSE,bool = TEST::value> struct gnx_if : ELSE {};

template <class TEST,class IF,class ELSE> struct gnx_if<TEST,IF,ELSE,true> : IF {};

//-------------------------------------------------------------------------------------gnx_enable_if
/* Provides a type metadata if the given boolean metadata is true (used for SFINAE). */
template <bool TEST,class TYPE> struct _gnx_enable_if_ {};

template <class TEST,class TYPE = void>
struct gnx_enable_if : _gnx_enable_if_<TEST::value,TYPE> {};

template <class TYPE> struct _gnx_enable_if_<true,TYPE> : gnx_type<TYPE> {};

//--------------------------------------------------------------------------------------------gnx_or
/* Returns the logical "or" operation between two boolean metadata. */
template <class TYPE1,class TYPE2,bool TEST> struct _gnx_or_if_;

template <class TYPE1,class TYPE2> struct gnx_or : _gnx_or_if_<TYPE1,TYPE2,TYPE1::value> {};

template <class TYPE1,class TYPE2> struct _gnx_or_if_<TYPE1,TYPE2,true> : gnx_true {};

template <class TYPE1,class TYPE2> struct _gnx_or_if_<TYPE1,TYPE2,false> : TYPE2 {};

//-------------------------------------------------------------------------------------------gnx_and
/* Returns the logical "and" operation between two boolean metadata. */
template <class TYPE1,class TYPE2,bool TEST> struct _gnx_and_if_;

template <class TYPE1,class TYPE2> struct gnx_and : _gnx_and_if_<TYPE1,TYPE2,TYPE1::value> {};

template <class TYPE1,class TYPE2> struct _gnx_and_if_<TYPE1,TYPE2,true> : TYPE2 {};

template <class TYPE1,class TYPE2> struct _gnx_and_if_<TYPE1,TYPE2,false> : gnx_false {};

//-------------------------------------------------------------------------------------------gnx_not
/* Returns the logical "not" operation applied on a boolean metadata. */
template <class TYPE> struct gnx_not : gnx_value<bool,!TYPE::value> {};

//----------------------------------------------------------------------------------------gnx_equals
/* Returns whether two metadata values are equal. */
template <class VALUE1,class VALUE2> struct gnx_equals
: gnx_value<bool,(VALUE1::value==VALUE2::value)> {};

//---------------------------------------------------------------------------------------gnx_greater
/* Returns whether a metadata value is greater than another one. */
template <class VALUE1,class VALUE2> struct gnx_greater
: gnx_value<bool,(VALUE1::value>VALUE2::value)> {};

//-------------------------------------------------------------------------------------------gnx_nil
/* Represents "no type". */
struct gnx_nil {};

//------------------------------------------------------------------------------------gnx_remove_ref
/* Removes the reference aspect from a type. */
template <class TYPE> struct gnx_remove_ref : gnx_type<TYPE> {};
template <class TYPE> struct gnx_remove_ref<TYPE &> : gnx_remove_ref<TYPE> {};

#ifdef gnx_cpp11
 template <class TYPE> struct gnx_remove_ref<TYPE &&> : gnx_remove_ref<TYPE> {};
#endif

//----------------------------------------------------------------------------------gnx_remove_const
/* Removes the constant aspect from a type. */
template <class TYPE> struct gnx_remove_const : gnx_type<TYPE> {};
template <class TYPE> struct gnx_remove_const<const TYPE> : gnx_remove_const<TYPE> {};

template <class TYPE> struct gnx_remove_const<TYPE &>
: gnx_type<typename gnx_remove_const<TYPE>::type &> {};

#ifdef gnx_cpp11
 template <class TYPE> struct gnx_remove_const<TYPE &&>
 : gnx_type<typename gnx_remove_const<TYPE>::type &&> {};
#endif

//-------------------------------------------------------------------------------------gnx_base_type
/* Returns the "base" of a type, meaning the type without reference and constant aspects. */
template <class TYPE> struct gnx_base_type
: gnx_remove_ref<typename gnx_remove_const<TYPE>::type> {};

//---------------------------------------------------------------------------------------gnx_forward
/* Helper function that allows "perfect forwarding". C++11 only. */
#ifdef gnx_cpp11
 template <class T>
 inline T && gnx_forward(typename gnx_remove_ref<T>::type & x)
 { return static_cast<T &&>(x); }
#endif

// Helpers //---------------------------------------------------------------------------------------
#ifdef gnx_cpp11
 template <class... PACK> using gnx_base_type_t    = typename gnx_base_type<PACK...>::type;
 template <class... PACK> using gnx_enable_if_t    = typename gnx_enable_if<PACK...>::type;
 template <class... PACK> using gnx_if_t           = typename gnx_if<PACK...>::type;
 template <class... PACK> using gnx_remove_const_t = typename gnx_remove_const<PACK...>::type;
 template <class... PACK> using gnx_remove_ref_t   = typename gnx_remove_ref<PACK...>::type;
#endif

// End //-------------------------------------------------------------------------------------------
#endif
